export declare type ChatMessageType = 'plaintext' | 'image' | 'video' | 'audio' | 'document' | 'location' | 'contact' | 'file' | 'url' | 'card';
export declare type IActingType = 'text' | 'voice' | 'video' | null;
export declare type ChatUser = {
    id: string;
    externalId: string;
    name: string;
    photoUrl: string;
    headline: string;
};
export interface ActingUser extends ChatUser {
    actingType?: IActingType;
}
export declare type ChannelMembership = {
    channelId: string;
    createdAt: string;
    id: string;
    listed: boolean;
    role: string;
    updatedAt: string;
    userId: string;
    externalRole?: string;
    watching: true;
    lastActingAt?: Date | null;
    actingType?: IActingType;
    lastReadMessageId?: number | null;
    lastRecievedMessageId?: number | null;
    user: ChatUser;
};
export declare type BlockedUser = {
    blockerId: string;
    blockedId: string;
};
export declare type ArchivedChannel = {
    userId: string;
    channelId: string;
};
export declare type DraftMessage = {
    userId?: string;
    channelId: string;
    draft: string;
};
export declare type ChatChannel = {
    id: string;
    externalId: string;
    name: string;
    photoUrl: string;
    channelType: 'direct' | 'group';
    lastMessage: ChatMessage | null;
    updatedAt: string;
    listed: boolean;
    watching: boolean;
    status: string;
    lastActivityAt: string;
    metadata?: Record<string, unknown>;
    unreadMessageCount?: number;
    isBlockingMe: boolean;
    isArchived: boolean;
    draft?: string;
    channelMemberships?: ChannelMembership[];
};
export declare type ChatChannelUpsert = {
    unreadMessageCount?: number;
    channelMembership?: ChannelMembership;
    channelMemberships?: ChannelMembership[];
};
export declare type ChatMessage = {
    id?: string;
    localId: string;
    channelExternalId?: string;
    channelId?: string;
    replyToId?: string | null;
    replyTo?: ChatMessage | null;
    messageType: ChatMessageType;
    content: string;
    createdAt?: string;
    updatedAt?: string;
    postedAt?: Date;
    serverReceivedAt?: string;
    deletedAt?: string;
    data?: Record<string, any>;
    senderId?: string;
    attachment?: {
        url: string;
        thumbnail?: string;
        size: number;
        contentType: string;
        height?: number;
        width?: number;
        blurhash?: string;
        length?: number;
    };
    isQueued?: boolean;
    isRetriable?: boolean;
    isDeleted?: boolean;
    file?: any;
    metadata?: Record<string, any>;
};
declare class ChatClient {
    private endpoint;
    private token;
    private channels;
    private socket;
    private eventRegistry;
    private users;
    private currentUser?;
    private isDisconnected;
    private lastJoinedChannelId;
    constructor(endpoint: string, token: string);
    listen(): void;
    listenToAppEvents(): void;
    connect(): void;
    joinChannel(channelId: string): Promise<unknown>;
    leaveChannel(channelId: string): void;
    sendMessage(chatMessage: ChatMessage, onBlockedHandler?: (message: string) => any): Promise<void>;
    triggerUserActing(lastActingAt: Date | null, channelId: string, actingType: IActingType): void;
    onUserActing(callback: (data: {
        channelId: string;
        userId: string;
        lastActingAt: Date | null;
        actingType: IActingType;
    }) => void): () => void;
    onConnect(callback: () => void): () => void;
    onReconnect(callback: () => void): () => void;
    onMaxRetryAttemptsReached(callback: () => void): () => void;
    onDisconnect(callback: () => void): () => void;
    onAllNewMessage(callback: (message: ChatMessage) => void): () => void;
    onNewMessage(channelId: string, callback: (message: ChatMessage) => void): () => void;
    onChannelUpdate(callback: (channel: ChatChannel) => void): () => void;
    onChannelMembershipsUpdate(callback: (channelMemberships: ChannelMembership[]) => void): () => void;
    onChannelUpsert(callback: (data: {
        channelId: string;
        upsert: ChatChannelUpsert;
    }) => void): () => void;
    onChannelCreate(callback: (channel: ChatChannel) => void): () => void;
    onChannelDelete(callback: (channel: ChatChannel) => void): () => void;
    onUserBlockChange(callback: (data: {
        userId: string;
        isBlocking: boolean;
    }) => void): () => void;
    onMessageDeleted(callback: (data: {
        channelId: string;
        messageId: string;
    }) => void): () => void;
    onUpdateDraft(callback: (data: {
        channelId: string;
        draft: string | undefined;
    }) => void): () => void;
    emitMarkReadUntil(channelId: string, messageId?: string): Promise<void>;
    getUnreadMessageCount(channel: ChatChannel): Promise<ChatChannel>;
    getAllUnreadMessageCount(): Promise<{
        unreadMessageCount: number;
    }>;
    loadEarlierChannels(earliestDate?: string, size?: number): Promise<ChatChannel[]>;
    loadLaterChannels(latestDate?: string, size?: number): Promise<ChatChannel[]>;
    getChannel(channelId?: string, channelExternalId?: string, filter?: {
        skipEmpty?: boolean;
        metadata?: Record<any, any>;
    }): Promise<ChatChannel[]>;
    getChannels(filter?: {
        skipEmpty?: boolean;
        metadata?: Record<string, string>;
        perPage?: number;
    }): Promise<ChatChannel[]>;
    loadMessagesBeforeId(channelId: string, messageIdBefore: number, size: number): Promise<ChatMessage[]>;
    loadMessagesAfterId(channelId: string, messageIdAfter: number, size: number): Promise<ChatMessage[]>;
    loadEarlierMessages(channelId: string, earliestDate?: string, size?: number): Promise<ChatMessage[]>;
    loadLaterMessages(channelId: string, latestDate?: string, size?: number): Promise<ChatMessage[]>;
    searchMessages(channelId: string, query: string, size?: number): Promise<ChatMessage[]>;
    getUser(id: string): Promise<ChatUser>;
    me(): Promise<ChatUser>;
    sendAttachment(chatMessage: ChatMessage, file: any, onUploadProgress?: (progress: number) => any): Promise<any>;
    sendMessageExternally(chatMessage: ChatMessage): Promise<void>;
    blockUser(blockedId: string): Promise<any>;
    unBlockUser(blockedId: string): Promise<any>;
    getBlockedUsers(): Promise<BlockedUser[]>;
    isBlockingMe(blockerId: string): Promise<any>;
    archiveChannel(channelId: string): Promise<any>;
    unArchiveChannel(channelId: string): Promise<any>;
    getArchivedChannels(): Promise<ArchivedChannel[]>;
    updateDraftMessage(channelId: string, draft: string): Promise<any>;
    getDraftMessages(): Promise<DraftMessage[]>;
    deleteMessage(messageId: string): Promise<any>;
    getStatus(): Promise<{
        host: string;
    }>;
    private executeHTTP;
}
export default ChatClient;
//# sourceMappingURL=ChatClient.d.ts.map