import { ChannelMessageStorageStateType } from './useChannelData';
import concat from 'lodash/concat';

import ChatClient, { ChatMessage } from './ChatClient';
import EventRegistry from './EventRegistry';
import InMemoryCollection from './InMemoryCollection';

export const maxSize = 100;

/*
  ChannelMessageStorage handle storing messages. It should cover below cases
  -  Receiving a new message via socket
  -  Loading previous message
  -  Loading missed message in case of diconnecting and reconnecting.

  Known issues:
  - Currently it is using createdAt date as marker to load previous/later messages.
    However it is not 100% reliable in case multiple messages had the exact same time. 
  - Throwing error and retrying in case of error
*/
class ChannelMessageStorage {
  private channelId: string;
  private db: any;
  private messagesCollection: InMemoryCollection<ChatMessage>;
  private liveUpdate: boolean;
  private notMergedMessages: ChatMessage[];
  private hasMoreEarlier: boolean;
  private eventRegistry: EventRegistry;
  private loadingEarlier: boolean;
  private loadingLater: boolean;
  public initialMessagesLoaded: boolean;

  constructor(channelId: string) {
    this.channelId = channelId;
    this.messagesCollection = new InMemoryCollection<ChatMessage>('id');
    this.eventRegistry = new EventRegistry();
    this.notMergedMessages = [];
    this.liveUpdate = false;
    this.hasMoreEarlier = false;
    this.loadingEarlier = false;
    this.loadingLater = false;
    this.initialMessagesLoaded = false;
  }

  setLiveUpdate(liveUpdate: boolean) {
    this.liveUpdate = liveUpdate;
  }

  upsert(messages: ChatMessage[]) {
    const len = messages.length;
    for (let i = 0; i < len; i++) {
      if (messages[i].localId) {
        // make sure queued messsage is removed
        this.messagesCollection.remove(messages[i].localId);
      }
      this.messagesCollection.upsert(messages[i]);
    }
  }

  addLiveMessage(newMessages: ChatMessage[]) {
    if (this.liveUpdate) {
      if (this.messagesCollection.hasChanges(newMessages)) {
        this.upsert(newMessages);
        this.eventRegistry.emit('update', {});
      }
    } else {
      const messages = concat(this.notMergedMessages, newMessages);
      this.notMergedMessages = messages;
    }
  }

  removeMessage(id: string) {
    this.messagesCollection.remove(id);
    this.eventRegistry.emit('update', {});
  }

  softDeleteMessage(id: string) {
    const message = this.messagesCollection.find(id);
    this.messagesCollection.upsert({
      ...message,
      isDeleted: true,
      content: 'This message was deleted',
    });
    this.eventRegistry.emit('update', {});
  }

  markAllQueuedMessagesAsRetriable() {
    const messages = this.sortedMessages();
    for (let i = 0; i < messages.length; i++) {
      if (messages[i].isQueued) {
        messages[i] = {
          ...messages[i],
          isRetriable: true,
        };
        // mark message as failed
        this.messagesCollection.upsert(messages[i]);
      }
    }
    if (this.liveUpdate) {
      this.eventRegistry.emit('update', {});
    }
  }

  addLaterMessage(laterMessages: ChatMessage[]) {
    this.upsert(laterMessages);
    if (maxSize && laterMessages.length < maxSize) {
      //TODO: remove duplicates from messagesCache
      this.upsert(this.notMergedMessages);
      this.notMergedMessages = [];
      this.liveUpdate = true;
    } else {
      this.liveUpdate = false;
    }
  }

  addEarlierMessage(ealierMessages: ChatMessage[]) {
    this.upsert(ealierMessages);
    if (maxSize && ealierMessages.length < maxSize) {
      this.hasMoreEarlier = false;
    } else {
      this.hasMoreEarlier = true;
    }
  }

  sortedMessages() {
    return this.messagesCollection.sort('createdAt');
  }

  latestDate() {
    const messages = this.messagesCollection.sort('updatedAt', 'desc');
    return messages[0]?.updatedAt;
  }

  earliestDate() {
    const messages = this.sortedMessages();
    return messages[0]?.createdAt;
  }

  state(): ChannelMessageStorageStateType {
    return {
      messages: this.sortedMessages(),
      liveUpdate: this.liveUpdate,
      hasMoreEarlier: this.hasMoreEarlier,
      loadingEarlier: this.loadingEarlier,
      loadingLater: this.loadingLater,
    };
  }

  async loadEarlier(
    chatClient: ChatClient,
    unreadMessageCount?: number,
    lastReadMessageId?: number
  ) {
    this.loadingEarlier = true;
    let earlierMessages;
    try {
      if (!unreadMessageCount || unreadMessageCount < maxSize) {
        this.liveUpdate = true;
      } else {
        this.liveUpdate = false;
      }
      if (unreadMessageCount) {
        const limitAfter = Math.min(unreadMessageCount + 1, maxSize);
        const limitBefore = Math.max(maxSize - limitAfter, 0);

        const messagesBeforeLastReadMessageId =
          limitBefore > 0
            ? await chatClient.loadMessagesBeforeId(
                this.channelId,
                lastReadMessageId || 0,
                limitBefore
              )
            : [];
        const messagesAfterLastReadMessageId =
          await chatClient.loadMessagesAfterId(
            this.channelId,
            lastReadMessageId || 0,
            limitAfter
          );
        earlierMessages = [
          ...messagesBeforeLastReadMessageId,
          ...messagesAfterLastReadMessageId,
        ];
      } else {
        earlierMessages = await chatClient.loadEarlierMessages(
          this.channelId,
          this.earliestDate(),
          maxSize
        );
      }
      this.initialMessagesLoaded = true;
      this.loadingEarlier = false;
      this.addEarlierMessage(earlierMessages);
      this.eventRegistry.emit('update', {});
    } catch (error) {
      //Emit Error
      console.log(error);
      this.loadingEarlier = false;
      this.eventRegistry.emit('update', {});
    }
  }

  async loadLater(chatClient: ChatClient) {
    this.loadingLater = true;
    this.eventRegistry.emit('update', {});
    try {
      const laterMessages = await chatClient.loadLaterMessages(
        this.channelId,
        this.latestDate(),
        maxSize
      );
      this.loadingLater = false;
      this.addLaterMessage(laterMessages);
      this.eventRegistry.emit('update', {});
    } catch (error) {
      this.loadingLater = false;
      this.eventRegistry.emit('update', {});
      throw error;
    }
  }

  async loadAllLater(chatClient: ChatClient) {
    try {
      do {
        await this.loadLater(chatClient);
      } while (!this.liveUpdate);
    } catch (error) {}
  }

  onUpdate(callback: () => void) {
    return this.eventRegistry.on('update', callback);
  }
}

export default ChannelMessageStorage;
