import { ChannelStoragStateType } from './useChannelsData';
import ChatClient, {
  ArchivedChannel,
  BlockedUser,
  ChatChannel,
  DraftMessage,
} from './ChatClient';
import EventRegistry from './EventRegistry';
import InMemoryCollection from './InMemoryCollection';

const maxSize = 100;
class ChannelsStorage {
  private eventRegistry: EventRegistry;
  private channelsCollection: InMemoryCollection<ChatChannel>;
  private loadingEarlier: boolean = false;
  private hasMoreEarlier: boolean = false;
  private errorMessage?: string;
  private liveUpdate: boolean;
  private loadingLater: boolean;
  private initialDataLoaded: boolean;
  public maxRetryAttemptsReached: boolean;
  private blackList: BlockedUser[];
  private archivedChannels: ArchivedChannel[];
  private draftMessages: DraftMessage[];

  constructor() {
    this.channelsCollection = new InMemoryCollection('id');
    this.eventRegistry = new EventRegistry();
    this.liveUpdate = true;
    this.loadingLater = false;
    this.initialDataLoaded = false;
    this.maxRetryAttemptsReached = false;
    this.blackList = [];
    this.archivedChannels = [];
    this.draftMessages = [];
  }

  upsert(channels: ChatChannel[]) {
    const len = channels.length;
    for (let i = 0; i < len; i++) {
      if (channels[i].lastMessage) {
        // only upsert unempty channels
        const { channelMemberships, ...rest } = channels[i];
        this.channelsCollection.upsert(rest);
      }
    }
  }

  update(channel: ChatChannel) {
    const { channelMemberships, ...rest } = channel;
    if (this.channelsCollection.hasChanges([rest])) {
      this.upsert([rest]);
      this.eventRegistry.emit('update', { channel: { ...rest } });
    }
  }

  remove(channelId: string) {
    this.channelsCollection.remove(channelId);
    this.eventRegistry.emit('delete', { channelId });
  }

  addLaterChannels(laterChannels: ChatChannel[]) {
    this.upsert(laterChannels);
    if (maxSize && laterChannels.length < maxSize) {
      this.liveUpdate = true;
    } else {
      this.liveUpdate = false;
    }
  }

  sortedChannels() {
    return this.channelsCollection.sort('lastActivityAt', 'desc');
  }

  earliestDate() {
    const channels = this.channelsCollection.sort('lastActivityAt', 'asc');
    return channels[0]?.lastActivityAt;
  }

  latestDate() {
    const channels = this.channelsCollection.sort('updatedAt', 'desc');
    return channels[0]?.updatedAt;
  }

  loadEarlier(chatClient: ChatClient): Promise<ChatChannel[]> {
    this.loadingEarlier = true;
    return new Promise((resolve) => {
      chatClient
        .loadEarlierChannels(this.earliestDate(), maxSize)
        .then((earlierChannels) => {
          this.loadingEarlier = false;
          this.upsert(earlierChannels);
          if (maxSize && earlierChannels.length < maxSize) {
            this.hasMoreEarlier = false;
          } else {
            this.hasMoreEarlier = true;
          }
          // TODO setLastRecievedMessageId(channelX.channelId, channelX.lastMessage.id)
          this.errorMessage = undefined;
          this.initialDataLoaded = true;
          this.eventRegistry.emit('update', {});
          resolve(earlierChannels);
        })
        .catch((error: any) => {
          //Emit Error
          this.loadingEarlier = false;
          this.errorMessage = error.message;
          this.eventRegistry.emit('update', {});
          resolve([]);
        });
    });
  }

  async loadLater(chatClient: ChatClient): Promise<ChatChannel[]> {
    this.loadingLater = true;
    this.eventRegistry.emit('update', {});
    try {
      const laterChannels = await chatClient.loadLaterChannels(
        this.latestDate(),
        maxSize
      );
      // TODO setLastRecievedMessageId(channelX.channelId, channelX.lastMessage.id)
      this.loadingLater = false;
      this.addLaterChannels(laterChannels);
      this.eventRegistry.emit('update', {});
      return laterChannels;
    } catch (error) {
      this.loadingLater = false;
      this.eventRegistry.emit('update', {});
      console.log(error);
      return [];
    }
  }

  async loadAllLater(chatClient: ChatClient): Promise<ChatChannel[]> {
    if (!this.initialDataLoaded) return [];
    this.liveUpdate = false;
    const allLaterChannels = [];
    try {
      while (!this.liveUpdate) {
        const channels = await this.loadLater(chatClient);
        allLaterChannels.push(...channels);
      }
      return allLaterChannels;
    } catch (error) {
      return allLaterChannels;
    }
  }

  find(channelId: string): ChatChannel | undefined {
    const doc = this.channelsCollection.find(channelId);
    return doc;
  }

  async get(
    channelId: string,
    chatClient: ChatClient,
    filter?: { skipEmpty?: boolean }
  ): Promise<ChatChannel> {
    const doc = this.channelsCollection.find(channelId);
    if (doc) return doc;
    const channels = await chatClient.getChannel(channelId, undefined, filter);
    if (channels.length > 0) {
      this.upsert(channels);
      return channels[0];
    }
    throw new Error('Channel not found');
  }

  async getByExternalId(
    channelExternalId: string,
    chatClient: ChatClient
  ): Promise<ChatChannel> {
    const doc = this.channelsCollection.findBy({
      externalId: channelExternalId,
    });
    if (doc) return doc;
    const channels = await chatClient.getChannel(undefined, channelExternalId);
    if (channels.length > 0) {
      this.upsert(channels);
      return channels[0];
    }
    throw new Error('Channel not found');
  }

  state(): ChannelStoragStateType {
    return {
      liveUpdate: this.liveUpdate,
      channels: this.sortedChannels(),
      hasMoreEarlier: this.hasMoreEarlier,
      loadingEarlier: this.loadingEarlier,
      loadingLater: this.loadingLater,
      errorMessage: this.errorMessage,
      maxRetryAttemptsReached: this.maxRetryAttemptsReached,
      blackList: this.blackList,
    };
  }

  onUpdate(callback: (data: any) => void) {
    return this.eventRegistry.on('update', callback);
  }

  onDelete(callback: (data: any) => void) {
    return this.eventRegistry.on('delete', callback);
  }

  setMaxRetryAttemptsReached(flag: boolean) {
    this.maxRetryAttemptsReached = flag;
    this.eventRegistry.emit('update', {});
  }

  setBlackList(blackList: BlockedUser[]) {
    this.blackList = blackList;
    this.eventRegistry.emit('update', {});
  }

  setArchivedChannels(
    archivedChannels: ArchivedChannel[],
    targetChannelId?: string
  ) {
    const channels = this.sortedChannels().map((channel) => ({
      ...channel,
      isArchived: archivedChannels.some((ac) => ac.channelId === channel.id),
    }));
    this.upsert(channels);
    this.eventRegistry.emit('update', {
      channel: channels.find((c) => c.id === targetChannelId),
    });
  }

  setDraftMessages(draftMessages: DraftMessage[], targetChannelId?: string) {
    const channels = this.sortedChannels().map((channel) => ({
      ...channel,
      draft: draftMessages.find((dm) => dm.channelId === channel.id)?.draft,
    }));
    this.upsert(channels);
    this.eventRegistry.emit('update', {
      channel: channels.find(
        (c) => c.id.toString() === targetChannelId?.toString()
      ),
    });
  }

  updateDraftMessage(channelId: string, draft: string | undefined) {
    const nDraftMessages = this.draftMessages.filter(
      (draftMessage) =>
        draftMessage.channelId.toString() === channelId.toString()
    );
    if (!draft) {
      return this.setDraftMessages(nDraftMessages);
    }
    let draftMessage = this.draftMessages.find(
      (draftMessage) =>
        draftMessage.channelId.toString() === channelId.toString()
    );
    if (!draftMessage) {
      draftMessage = { channelId, draft };
    } else {
      draftMessage.draft = draft;
    }
    return this.setDraftMessages([...nDraftMessages, draftMessage]);
  }
}

export default ChannelsStorage;
