import * as qs from 'qs';
import { io as SocketIoClient } from 'socket.io-client';
import axios, { AxiosRequestConfig } from 'axios';

import EventRegistry from './EventRegistry';
import packageJson from '../package.json';
import {
  serializeChannel,
  serializeChannelMembership,
  serializeChannelMemberships,
  serializeChannels,
  serializeMessage,
  serializeMessages,
} from './serializers';

export type ChatMessageType =
  | 'plaintext'
  | 'image'
  | 'video'
  | 'audio'
  | 'document'
  | 'location'
  | 'contact'
  | 'file'
  | 'url'
  | 'card';

export type IActingType = 'text' | 'voice' | 'video' | null;

export type ChatUser = {
  id: string;
  externalId: string;
  name: string;
  photoUrl: string;
  headline: string;
};

export interface ActingUser extends ChatUser {
  actingType?: IActingType;
}

export type ChannelMembership = {
  channelId: string;
  createdAt: string;
  id: string;
  listed: boolean;
  role: string;
  updatedAt: string;
  userId: string;
  externalRole?: string;
  watching: true;
  lastActingAt?: Date | null;
  actingType?: IActingType;
  lastReadMessageId?: number | null;
  lastRecievedMessageId?: number | null;
  user: ChatUser;
};

export type BlockedUser = {
  blockerId: string;
  blockedId: string;
};

export type ArchivedChannel = {
  userId: string;
  channelId: string;
};
export type DraftMessage = {
  userId?: string;
  channelId: string;
  draft: string;
};

export type ChatChannel = {
  id: string;
  externalId: string;
  name: string;
  photoUrl: string;
  channelType: 'direct' | 'group';
  lastMessage: ChatMessage | null;
  updatedAt: string;
  listed: boolean;
  watching: boolean;
  status: string;
  lastActivityAt: string;
  metadata?: Record<string, unknown>;
  unreadMessageCount?: number;
  isBlockingMe: boolean;
  isArchived: boolean;
  draft?: string;
  channelMemberships?: ChannelMembership[];
};

export type ChatChannelUpsert = {
  unreadMessageCount?: number;
  channelMembership?: ChannelMembership;
  channelMemberships?: ChannelMembership[];
};

export type ChatMessage = {
  id?: string;
  localId: string;
  channelExternalId?: string;
  channelId?: string;
  replyToId?: string | null;
  replyTo?: ChatMessage | null;
  messageType: ChatMessageType;
  content: string;
  createdAt?: string;
  updatedAt?: string;
  postedAt?: Date;
  serverReceivedAt?: string;
  deletedAt?: string;
  data?: Record<string, any>;
  senderId?: string;
  attachment?: {
    url: string;
    thumbnail?: string;
    size: number;
    contentType: string;
    height?: number;
    width?: number;
    blurhash?: string;
    length?: number;
  };
  isQueued?: boolean;
  isRetriable?: boolean;
  isDeleted?: boolean;
  file?: any;
  metadata?: Record<string, any>;
};

const MAX_RETRY_ATTEMPS = 10;

class ChatClient {
  private endpoint: string = 'http://localhost:4000';

  private token: string = '';

  private channels: ChatChannel[] = [];

  private socket: ReturnType<typeof SocketIoClient>;

  private eventRegistry = new EventRegistry();

  private users: { [id: string]: ChatUser } = {};

  private currentUser?: ChatUser;

  private isDisconnected: boolean = false;

  private lastJoinedChannelId: string | null = null;

  constructor(endpoint: string, token: string) {
    this.endpoint = endpoint;
    this.token = token;
    // TODO dont connect just initialize
    this.socket = SocketIoClient(this.endpoint, {
      autoConnect: false,
      forceNew: true,
      transports: ['websocket'],
      query: {
        token: this.token,
      },
    });
  }

  listen() {
    this.listenToAppEvents();

    this.socket.on('connect', () => {
      if (this.isDisconnected) {
        setTimeout(async () => {
          if (this.lastJoinedChannelId) {
            await this.joinChannel(this.lastJoinedChannelId);
          }
          this.eventRegistry.emit('reconnect', {});
        }, 5000);

        this.isDisconnected = false;
      }

      this.eventRegistry.emit('connect', {});
    });

    this.socket.on('connect_error', (e: any) => {
      // ...
    });

    this.socket.on('disconnect', (data: any) => {
      this.isDisconnected = true;
    });

    // Listen on IO events

    this.socket.io.on('error', (e: any) => {
      // ...
    });

    this.socket.io.on('reconnect_attempt', (attempt: any) => {
      if (attempt === MAX_RETRY_ATTEMPS) {
        this.eventRegistry.emit('maxRetryAttemptsReached', {});
      }
    });

    this.socket.io.on('reconnect_error', (e: any) => {
      // ...
    });

    this.socket.io.on('reconnect_failed', () => {
      // ...
    });

    this.socket.io.on('ping', () => {
      // ...
    });
  }

  listenToAppEvents() {
    this.socket.on('message', (data: ChatMessage) => {
      this.eventRegistry.emit('message', data ? serializeMessage(data) : null);
    });

    this.socket.on(
      'acting',
      async ({
        userId,
        channelId,
        lastActingAt,
        actingType,
      }: {
        userId: string;
        channelId: string;
        lastActingAt: Date;
        actingType: IActingType;
      }) => {
        this.eventRegistry.emit('acting', {
          userId,
          channelId,
          lastActingAt,
          actingType,
        });
      }
    );

    this.socket.on(
      'channelUpdate',
      (data: ChatChannel & { channelMemberships?: ChannelMembership[] }) => {
        if (data.channelMemberships) {
          this.eventRegistry.emit(
            'channelMembershipsUpdate',
            serializeChannelMemberships(data.channelMemberships)
          );
        }
        const { channelMemberships, ...channel } = data;
        this.eventRegistry.emit('channelUpdate', serializeChannel(channel));
      }
    );
    this.socket.on(
      'channelUpsert',
      (data: { channelId: string; upsert: ChatChannelUpsert }) => {
        this.eventRegistry.emit('channelUpsert', {
          channelId: String(data.channelId),
          upsert: {
            unreadMessageCount:
              data.upsert.unreadMessageCount !== undefined
                ? parseInt(String(data.upsert.unreadMessageCount), 10)
                : undefined,
            channelMembership: data.upsert.channelMembership
              ? serializeChannelMembership(data.upsert.channelMembership)
              : undefined,
            channelMemberships: data.upsert.channelMemberships
              ? serializeChannelMemberships(data.upsert.channelMemberships)
              : undefined,
          },
        });
      }
    );
    this.socket.on('channelDelete', (data: any) => {
      this.eventRegistry.emit('channelDelete', data);
    });
    this.socket.on('channelCreate', (data: any) => {
      this.eventRegistry.emit('channelCreate', data);
    });
    this.socket.on(
      'markedAllAsRead',
      (data: {
        channelId: string;
        updatedChannelMembership: ChannelMembership;
      }) => {
        const upsert: ChatChannelUpsert = {
          unreadMessageCount: 0,
          channelMembership: serializeChannelMembership(
            data.updatedChannelMembership
          ),
        };
        this.eventRegistry.emit('channelUpsert', {
          channelId: String(data.channelId),
          upsert,
        });
      }
    );

    this.socket.on('userBlockChange', (data: any) => {
      this.eventRegistry.emit('userBlockChange', data);
    });

    this.socket.on('messageDeleted', (data: any) => {
      this.eventRegistry.emit('messageDeleted', data);
    });
    this.socket.on('updateDraft', (data: any) => {
      this.eventRegistry.emit('updateDraft', data);
    });
  }

  connect() {
    this.listen();
    this.socket.open();
  }

  joinChannel(channelId: string) {
    return new Promise((resolve) => {
      this.lastJoinedChannelId = channelId;
      this.socket.emit('join-channel', { id: channelId }, resolve);
    });
  }

  leaveChannel(channelId: string) {
    this.lastJoinedChannelId = null;
    this.socket.emit('leave-channel', { id: channelId });
  }

  sendMessage(
    chatMessage: ChatMessage,
    onBlockedHandler?: (message: string) => any
  ): Promise<void> {
    return new Promise((resolve, reject) => {
      this.socket.emit(
        'message',
        chatMessage,
        (status: Number, message: string) => {
          if (status === 403) {
            return onBlockedHandler?.(message);
          }
          resolve();
        }
      );
      // when to reject? timeout maybe
    });
  }

  triggerUserActing(
    lastActingAt: Date | null,
    channelId: string,
    actingType: IActingType
  ) {
    this.socket.emit('acting', {
      id: channelId,
      lastActingAt,
      actingType,
    });
  }

  onUserActing(
    callback: (data: {
      channelId: string;
      userId: string;
      lastActingAt: Date | null;
      actingType: IActingType;
    }) => void
  ) {
    return this.eventRegistry.on('acting', callback);
  }

  onConnect(callback: () => void): () => void {
    return this.eventRegistry.on('connect', callback);
  }

  onReconnect(callback: () => void): () => void {
    return this.eventRegistry.on('reconnect', callback);
  }

  onMaxRetryAttemptsReached(callback: () => void): () => void {
    return this.eventRegistry.on('maxRetryAttemptsReached', callback);
  }

  onDisconnect(callback: () => void): () => void {
    return this.eventRegistry.on('disconnect', callback);
  }

  onAllNewMessage(callback: (message: ChatMessage) => void): () => void {
    return this.eventRegistry.on('message', callback);
  }

  onNewMessage(
    channelId: string,
    callback: (message: ChatMessage) => void
  ): () => void {
    return this.eventRegistry.on(
      'message',
      callback,
      (message: ChatMessage) => message.channelId === channelId
    );
  }

  onChannelUpdate(callback: (channel: ChatChannel) => void) {
    return this.eventRegistry.on('channelUpdate', callback);
  }
  onChannelMembershipsUpdate(
    callback: (channelMemberships: ChannelMembership[]) => void
  ) {
    return this.eventRegistry.on('channelMembershipsUpdate', callback);
  }

  onChannelUpsert(
    callback: (data: { channelId: string; upsert: ChatChannelUpsert }) => void
  ) {
    return this.eventRegistry.on('channelUpsert', callback);
  }

  onChannelCreate(callback: (channel: ChatChannel) => void) {
    return this.eventRegistry.on('channelCreate', callback);
  }

  onChannelDelete(callback: (channel: ChatChannel) => void) {
    return this.eventRegistry.on('channelDelete', callback);
  }

  onUserBlockChange(
    callback: (data: { userId: string; isBlocking: boolean }) => void
  ) {
    return this.eventRegistry.on('userBlockChange', callback);
  }

  onMessageDeleted(
    callback: (data: { channelId: string; messageId: string }) => void
  ) {
    return this.eventRegistry.on('messageDeleted', callback);
  }

  onUpdateDraft(
    callback: (data: { channelId: string; draft: string | undefined }) => void
  ) {
    return this.eventRegistry.on('updateDraft', callback);
  }

  async emitMarkReadUntil(channelId: string, messageId?: string) {
    this.socket.emit(
      'markReadUntil',
      { channelId, messageId },
      (data: { updatedChannelMembership: any; unreadMessageCount: number }) => {
        const upsert: ChatChannelUpsert = {
          unreadMessageCount: parseInt(
            String(data.unreadMessageCount || 0),
            10
          ),
          channelMembership: serializeChannelMembership(
            data.updatedChannelMembership
          ),
        };
        this.eventRegistry.emit('channelUpsert', {
          channelId,
          upsert,
        });
      }
    );
  }

  async getUnreadMessageCount(channel: ChatChannel): Promise<ChatChannel> {
    try {
      const { count: unreadMessageCount } = await this.executeHTTP(
        'GET',
        `/api/client/chat/channels/unread-count/${channel.id}`,
        {},
        {}
      );
      return { ...channel, unreadMessageCount };
    } catch (error: any) {
      console.log(error.message);
      return channel;
    }
  }

  async getAllUnreadMessageCount(): Promise<{ unreadMessageCount: number }> {
    const { count: unreadMessageCount } = await this.executeHTTP(
      'GET',
      `/api/client/chat/channels/unread-count`,
      {},
      {}
    );
    return { unreadMessageCount };
  }

  async loadEarlierChannels(earliestDate?: string, size: number = 20) {
    const params: any = {
      filter: { skipEmpty: true, listed: true },
      perPage: size,
    };
    if (earliestDate) params.filter['lastActivityBefore'] = earliestDate;
    const channels: ChatChannel[] = await this.executeHTTP(
      'GET',
      '/api/client/chat/channels',
      params,
      {}
    );
    return serializeChannels(channels);
  }

  async loadLaterChannels(latestDate?: string, size: number = 20) {
    const params: any = {
      filter: { skipEmpty: true, listed: true },
      perPage: size,
    };
    if (latestDate) params.filter['updatedAfter'] = latestDate;
    const channels: ChatChannel[] = await this.executeHTTP(
      'GET',
      '/api/client/chat/channels',
      params,
      {}
    );
    return serializeChannels(channels);
  }

  async getChannel(
    channelId?: string,
    channelExternalId?: string,
    filter?: { skipEmpty?: boolean; metadata?: Record<any, any> }
  ): Promise<ChatChannel[]> {
    const filters: any = { listed: false };
    if (channelId || channelExternalId) {
      if (channelId) filters['id'] = channelId;
      if (channelExternalId) filters['externalId'] = channelExternalId;
      if (filter?.metadata)
        filters['metadata'] = JSON.stringify(filter.metadata);
      const channels: ChatChannel[] = await this.executeHTTP(
        'GET',
        '/api/client/chat/channels',
        { filter: { ...filters, ...filter } },
        {}
      );

      return serializeChannels(channels);
    } else {
      return [];
    }
  }
  async getChannels(filter?: {
    skipEmpty?: boolean;
    metadata?: Record<string, string>;
    perPage?: number;
  }): Promise<ChatChannel[]> {
    try {
      const filters: any = { listed: false };
      if (filter?.metadata)
        filters['metadata'] = JSON.stringify(filter.metadata);
      const channels: ChatChannel[] = await this.executeHTTP(
        'GET',
        '/api/client/chat/channels',
        { filter: { ...filters, ...filter }, perPage: filter?.perPage || 100 },
        {}
      );
      return serializeChannels(channels);
    } catch (error) {
      console.log(error);
      return [];
    }
  }

  async loadMessagesBeforeId(
    channelId: string,
    messageIdBefore: number,
    size: number
  ): Promise<ChatMessage[]> {
    const messages = await this.executeHTTP(
      'GET',
      '/api/client/chat/messages',
      {
        filter: {
          channelId,
          messageIdBefore,
        },
        perPage: size,
      },
      {}
    );
    return serializeMessages(messages);
  }

  async loadMessagesAfterId(
    channelId: string,
    messageIdAfter: number,
    size: number
  ): Promise<ChatMessage[]> {
    const messages = await this.executeHTTP(
      'GET',
      '/api/client/chat/messages',
      {
        filter: {
          channelId,
          messageIdAfter,
        },
        perPage: size,
      },
      {}
    );
    return serializeMessages(messages);
  }

  async loadEarlierMessages(
    channelId: string,
    earliestDate?: string,
    size: number = 20
  ): Promise<ChatMessage[]> {
    const params: any = { filter: { channelId }, perPage: size };

    if (earliestDate) params.filter['createdBefore'] = earliestDate;
    const messages = await this.executeHTTP(
      'GET',
      '/api/client/chat/messages',
      params,
      {}
    );
    return serializeMessages(messages);
  }

  async loadLaterMessages(
    channelId: string,
    latestDate?: string,
    size: number = 20
  ) {
    const params: any = { filter: { channelId }, perPage: size };
    if (latestDate) params.filter['updatedAfter'] = latestDate;
    const messages = await this.executeHTTP(
      'GET',
      '/api/client/chat/messages',
      params,
      {}
    );
    return serializeMessages(messages);
  }

  async searchMessages(channelId: string, query: string, size: number = 100) {
    const params: any = { filter: { channelId }, perPage: size };
    if (query) params.filter['query'] = query;
    const messages = await this.executeHTTP(
      'GET',
      '/api/client/chat/messages',
      params,
      {}
    );
    return serializeMessages(messages);
  }

  async getUser(id: string) {
    //check internal cache otherwise fetch from API
    if (this.users[id]) {
      return this.users[id];
    }
    const users: ChatUser[] = await this.executeHTTP(
      'GET',
      '/api/client/chat/users',
      { ids: [id] },
      {}
    );
    users.forEach((user) => {
      this.users[user.id] = user;
    });
    return this.users[id];
  }

  async me() {
    if (!this.currentUser)
      this.currentUser = await this.executeHTTP(
        'GET',
        '/api/client/chat/me',
        {},
        {}
      );
    return this.currentUser!;
  }

  async sendAttachment(
    chatMessage: ChatMessage,
    file: any,
    onUploadProgress?: (progress: number) => any
  ) {
    var body = new FormData();
    body.append('attachment', file);
    return this.executeHTTP(
      'POST',
      '/api/client/chat/sendMessage',
      { ...chatMessage },
      body,
      onUploadProgress
    );
  }

  sendMessageExternally(chatMessage: ChatMessage): Promise<void> {
    return this.executeHTTP(
      'POST',
      '/api/client/chat/sendMessage',
      { ...chatMessage },
      {}
    );
  }

  async blockUser(blockedId: string) {
    return this.executeHTTP(
      'PUT',
      `/api/client/chat/users/block/${blockedId}`,
      {},
      {}
    );
  }

  async unBlockUser(blockedId: string) {
    return this.executeHTTP(
      'PUT',
      `/api/client/chat/users/unblock/${blockedId}`,
      {},
      {}
    );
  }

  async getBlockedUsers(): Promise<BlockedUser[]> {
    return this.executeHTTP('GET', '/api/client/chat/users/blocked', {}, {});
  }

  async isBlockingMe(blockerId: string) {
    return this.executeHTTP(
      'GET',
      `/api/client/chat/users/isblockingme/${blockerId}`,
      {},
      {}
    );
  }

  async archiveChannel(channelId: string) {
    return this.executeHTTP(
      'PUT',
      `/api/client/chat/channels/archive/${channelId}`,
      {},
      {}
    );
  }

  async unArchiveChannel(channelId: string) {
    return this.executeHTTP(
      'PUT',
      `/api/client/chat/channels/unarchive/${channelId}`,
      {},
      {}
    );
  }

  async getArchivedChannels(): Promise<ArchivedChannel[]> {
    return this.executeHTTP(
      'GET',
      '/api/client/chat/channels/archived',
      {},
      {}
    );
  }

  async updateDraftMessage(channelId: string, draft: string) {
    return this.executeHTTP(
      'PUT',
      `/api/client/chat/users/drafts/${channelId}`,
      { draft },
      {}
    );
  }

  async getDraftMessages(): Promise<DraftMessage[]> {
    return this.executeHTTP('GET', '/api/client/chat/users/drafts', {}, {});
  }

  async deleteMessage(messageId: string): Promise<any> {
    return this.executeHTTP(
      'DELETE',
      `/api/client/chat/messages/${messageId}`,
      {},
      {}
    );
  }

  async getStatus(): Promise<{ host: string }> {
    return this.executeHTTP('GET', '/status', {}, {});
  }

  private async executeHTTP(
    method: 'GET' | 'POST' | 'PUT' | 'DELETE',
    path: string,
    urlParams: any,
    body: any,
    onUploadProgress?: (progress: number) => any
  ) {
    let queryString = '';
    if (urlParams) {
      queryString = qs.stringify(
        { ...urlParams, sdkVersion: packageJson.version },
        { arrayFormat: 'brackets' }
      );
    }
    const url = `${this.endpoint}${path}?${queryString}`;
    const options: AxiosRequestConfig<any> = {
      method,
      headers: {
        Authorization: `Bearer ${this.token}`,
      },
      onUploadProgress: (p) => {
        onUploadProgress?.(p.loaded / p.total);
      },
    };
    if (method === 'POST' || method === 'PUT') {
      options.data = body;
    }
    const response = await axios(url, options);
    if (response.status < 200 || response.status > 399) {
      throw new Error(response.statusText);
    }
    return response.data;
  }
}

export default ChatClient;
