import { diffInSeconds } from './utils';
import { v4 as uuidv4 } from 'uuid';
import flatten from 'lodash/flatten';

/* eslint-disable no-undef */
import ChannelMessageStorage from './ChannelMessageStorage';
import ChannelsStorage from './ChannelStorage';
import ChatClient, {
  ChatMessage,
  ChatMessageType,
  ChatChannel,
  ChatUser,
  IActingType,
  ChannelMembership,
} from './ChatClient';
import ChatMemoryStorageAdapter from './ChatMemoryStorageAdapter';
import ChannelMembershipsStorage from './ChannelMembershipStorage';

const DEFAULT_ENDPOINT = 'http://localhost:4000';

class ChatService {
  private chatMemoryStorageAdapter: ChatMemoryStorageAdapter;
  private chatClient: ChatClient;
  private channelsStorage: ChannelsStorage;
  private channelMembershipsStorage: ChannelMembershipsStorage;
  private lastActingAt: Date | null;
  private lastActingType: IActingType;
  private currentActiveChannel: ChatChannel | null;
  private static chatService: ChatService;
  private static token: string;
  public currentUser: ChatUser | null;

  static getService(token: string, endpoint?: string) {
    if (this.chatService && token === this.token) {
      return this.chatService;
    }
    if (this.chatService) {
      //TODO: disconnect and cleanup
    }
    this.token = token;
    this.chatService = new ChatService(endpoint || DEFAULT_ENDPOINT, token);
    return this.chatService;
  }

  constructor(endpoint: string, token: string) {
    this.chatMemoryStorageAdapter = new ChatMemoryStorageAdapter();
    this.chatClient = new ChatClient(endpoint, token);
    this.channelsStorage = new ChannelsStorage();
    this.channelMembershipsStorage = new ChannelMembershipsStorage();
    this.lastActingType = null;
    this.lastActingAt = null;
    this.currentActiveChannel = null;
    this.currentUser = null;

    this.chatClient.me().then((user) => {
      this.currentUser = user;
    });

    this.chatClient.onAllNewMessage(async (message: ChatMessage) => {
      if (message.channelId) {
        const channelMessageStorage = this.channelMessageStorage(
          message.channelId
        );
        if (
          channelMessageStorage.initialMessagesLoaded &&
          this.currentActiveChannel?.id !== message.channelId
        ) {
          channelMessageStorage.addLiveMessage([message]);
        }
      }
    });

    this.chatClient.onChannelUpdate(async (channel: ChatChannel) => {
      if (channel) {
        const currentUser = await this.getCurrentUser();
        const oldChannel = await this.channelsStorage.get(
          channel.id,
          this.chatClient
        );
        let updatedChannel = { ...oldChannel, ...channel };
        if (
          this.currentActiveChannel?.id !== channel.id &&
          channel.lastMessage?.senderId !== currentUser?.id
        ) {
          // if the channel is open then mark this message as read
          updatedChannel = await this.chatClient.getUnreadMessageCount(channel);
        }
        if (updatedChannel.channelMemberships) {
          this.channelMembershipsStorage.update(
            updatedChannel.channelMemberships
          );
          delete updatedChannel.channelMemberships;
        }
        this.channelsStorage.update(updatedChannel);
      }
    });

    this.chatClient.onChannelUpsert(async (data) => {
      let channel;
      try {
        channel = await this.channelsStorage.get(
          data.channelId,
          this.chatClient,
          { skipEmpty: true }
        );
      } catch (error) {
        // channel not found
        console.log('channel not found');
        return;
      }
      const { channelMembership, channelMemberships, ...rest } = data.upsert;
      const nChannel: ChatChannel = {
        ...channel,
        ...rest,
      };
      if (channelMemberships) {
        this.channelMembershipsStorage.update(channelMemberships);
      }
      if (channelMembership) {
        this.channelMembershipsStorage.update([channelMembership]);
      }
      this.channelsStorage.update(nChannel);
    });

    this.chatClient.onChannelMembershipsUpdate((channelMemberships) => {
      this.channelMembershipsStorage.upsert(channelMemberships);
    });

    this.chatClient.onChannelCreate((channel: ChatChannel) => {
      if (channel && channel.lastMessage) this.channelsStorage.update(channel);
    });

    this.chatClient.onChannelDelete((channel: ChatChannel) => {
      if (channel) this.channelsStorage.remove(channel.id);
    });

    this.chatClient.onConnect(() => {
      //TODO for all listed and watched channels load missed messages
    });

    this.chatClient.onReconnect(() => {
      this.channelsStorage.setMaxRetryAttemptsReached(false);
      this.channelsStorage.sortedChannels().forEach((channel) => {
        this.channelMessageStorage(channel.id)
          .sortedMessages()
          .filter((m) => m.isQueued)
          .forEach((message) => {
            this.retryMessage(message);
          });
      });
    });

    this.chatClient.onMaxRetryAttemptsReached(() => {
      this.channelsStorage.sortedChannels().forEach((channel) => {
        this.channelMessageStorage(
          channel.id
        ).markAllQueuedMessagesAsRetriable();
      });
      this.channelsStorage.setMaxRetryAttemptsReached(true);
    });

    this.chatClient.onUserActing(
      async ({ channelId, userId, lastActingAt, actingType }) => {
        const channelMemberships =
          this.channelMembershipsStorage.findMany(channelId);
        if (channelMemberships) {
          const nChannelMemberships = channelMemberships.map(
            (channelMemberShip) => {
              if (channelMemberShip.userId === userId) {
                return {
                  ...channelMemberShip,
                  lastActingAt,
                  actingType,
                };
              }
              return { ...channelMemberShip };
            }
          );
          this.channelMembershipsStorage.update(nChannelMemberships);
        } else {
          console.warn('No memberships found for channel', channelId);
        }
      }
    );

    this.chatClient.onUserBlockChange(({ userId, isBlocking }) => {
      this.channelsStorage.sortedChannels().forEach(async (channel) => {
        const channelMemberships =
          await this.channelMembershipsStorage.findMany(channel.id);
        if (channelMemberships?.some((cm) => cm.userId === userId)) {
          this.channelsStorage.update({ ...channel, isBlockingMe: isBlocking });
        }
      });
    });

    this.chatClient.onMessageDeleted(({ channelId, messageId }) => {
      this.channelMessageStorage(channelId).softDeleteMessage(messageId);
    });

    this.chatClient.onUpdateDraft(({ channelId, draft }) => {
      this.channelsStorage.updateDraftMessage(channelId, draft);
    });
  }

  start() {
    //TODO: prevent double connect
    this.chatClient.connect();
  }

  setCurrentActiveChannel(channel: ChatChannel | null) {
    this.currentActiveChannel = channel;
  }

  getCurrentActiveChannel(): ChatChannel | null {
    return this.currentActiveChannel;
  }

  channelMessageStorage(channelId: string): ChannelMessageStorage {
    return this.chatMemoryStorageAdapter.channelMessageStorage(channelId);
  }

  getChannelStorage(): ChannelsStorage {
    return this.channelsStorage;
  }

  getChannelMembershipsStorage(): ChannelMembershipsStorage {
    return this.channelMembershipsStorage;
  }

  async getCurrentUser(): Promise<ChatUser> {
    this.currentUser = await this.chatClient.me();
    return this.currentUser;
  }

  async getChannel(channelId: string) {
    const channel = await this.channelsStorage.get(channelId, this.chatClient);
    if (channel.channelMemberships) {
      this.channelMembershipsStorage.update(channel.channelMemberships);
    }
    return channel;
  }

  async getChannelByExternalId(channelExternalId: string) {
    return await this.channelsStorage.getByExternalId(
      channelExternalId,
      this.chatClient
    );
  }

  subscribeChannel(channelId: string): CallableFunction {
    const unsubscribeList: CallableFunction[] = [];
    let unsubscribed = false;
    const channelMessageStorage = this.channelMessageStorage(channelId);
    this.channelsStorage
      .get(channelId, this.chatClient)
      .then(async (channel) => {
        if (unsubscribed) {
          return;
        }
        const currentUser = await this.getCurrentUser();

        //Check if we need to join channel socket or the user channel is enough
        //FIXME channel does not have listed & watching. Those property are from ChannelMembership.
        if (channel.listed || channel.watching) {
          if (!channelMessageStorage.latestDate()) {
            //first time opening the channel. Load some messages
            channelMessageStorage.loadEarlier(this.chatClient);
          }
          return () => {};
        }

        this.chatClient.joinChannel(channel.id);

        unsubscribeList.push(() => {
          this.chatClient.leaveChannel(channel.id);
        });

        //if channel was previously opened and some messages are loaded then load the reset
        if (
          channelMessageStorage.latestDate() &&
          !channelMessageStorage.initialMessagesLoaded
        ) {
          channelMessageStorage.loadAllLater(this.chatClient);
        } else {
          const channelMemberships =
            await this.channelMembershipsStorage.findMany(channel.id);
          const lastReadMessageId =
            channelMemberships?.find((cm) => cm.userId === currentUser?.id)
              ?.lastReadMessageId || 0;
          channelMessageStorage.loadEarlier(
            this.chatClient,
            channel.unreadMessageCount,
            lastReadMessageId
          );
        }

        // update black list and is blocking me with latest
        const blackList = await this.chatClient.getBlockedUsers();
        this.channelsStorage.setBlackList(blackList);

        const channelMemberships =
          await this.channelMembershipsStorage.findMany(channel.id);
        const otherUserId = channelMemberships?.find(
          (cm) => cm.userId !== currentUser?.id
        )?.userId;
        if (otherUserId) {
          const isBlockingMe = await this.chatClient.isBlockingMe(otherUserId);

          this.channelsStorage.sortedChannels().forEach(async (channel) => {
            const channelMemberships =
              await this.channelMembershipsStorage.findMany(channel.id);
            if (channelMemberships?.some((cm) => cm.userId === otherUserId)) {
              this.channelsStorage.update({ ...channel, isBlockingMe });
            }
          });
        }

        const removeOnNewMessage = this.chatClient.onNewMessage(
          channelId,
          (message) => {
            channelMessageStorage.addLiveMessage([message]);

            // TODO setLastRecievedMessageId(channelId, message.id)
            if (
              message.senderId !== currentUser?.id &&
              channelId === this.currentActiveChannel?.id
            ) {
              console.log('marking all as read');
              this.markAllAsRead(channelId);
            }
          }
        );
        const removeOnConnect = this.chatClient.onConnect(() => {
          channelMessageStorage.loadAllLater(this.chatClient);
        });
        unsubscribeList.push(removeOnNewMessage);
        unsubscribeList.push(removeOnConnect);
      })
      .catch((_error) => {
        //throw error
      });

    return () => {
      // Following race condition may happens:
      // user subscrube to channel and we return the unsubscribe function
      // use unsubscribe before functions are pushed into unsubscribed list
      // In this case user will not be unsubscribed from OnNewMessage & OnConnect
      // To prevent this race condition we created unsubscribed flag
      unsubscribed = true;
      for (var i = 0; i < unsubscribeList.length; i++) {
        unsubscribeList[i]();
      }
    };
  }

  updateChannelMemberships(channelMemberships: ChannelMembership[]) {
    this.channelMembershipsStorage.update(channelMemberships);
  }

  subscribeChannelsUpdate() {
    const removeOnConnect = this.chatClient.onConnect(async () => {
      const channels = await this.channelsStorage.loadAllLater(this.chatClient);
      this.updateChannelMemberships(
        flatten(channels.map((channel) => channel.channelMemberships || []))
      );
    });

    if (!this.channelsStorage.earliestDate()) {
      this.channelsStorage.loadEarlier(this.chatClient).then((channels) => {
        this.updateChannelMemberships(
          flatten(channels.map((channel) => channel.channelMemberships || []))
        );
      });
    }
    return () => {
      removeOnConnect();
    };
  }

  async sendMessage(
    channelId: string,
    messageType: ChatMessageType,
    content: string,
    metadata?: Record<string, unknown>,
    replyTo?: ChatMessage | null,
    localId?: string
  ): Promise<void> {
    const currentUser = await this.getCurrentUser();
    // validate messageType
    const chatMessage: ChatMessage = {
      localId: localId || uuidv4(),
      channelId,
      messageType,
      content,
      postedAt: new Date(),
      createdAt: new Date().toString(),
      senderId: currentUser?.id,
      metadata,
      replyToId: replyTo?.id || null,
      replyTo,
    };
    // store the message as queued for retries
    this.channelMessageStorage(channelId).addLiveMessage([
      {
        ...chatMessage,
        id: chatMessage.localId,
        isQueued: true,
        isRetriable: this.channelsStorage.maxRetryAttemptsReached,
      },
    ]);

    return this.chatClient.sendMessage(chatMessage, (onBlockedMessage) => {
      this.channelMessageStorage(channelId).removeMessage(chatMessage.localId);
    });
  }

  retryMessage(
    message: ChatMessage,
    onUploadProgress?: (progress: number) => any
  ) {
    if (message.file) {
      return this.chatClient.sendAttachment(
        message,
        message.file,
        onUploadProgress
      );
    }
    return this.chatClient.sendMessage(message);
  }

  async triggerUserTyping(isTyping: boolean, channelExternalId: string) {
    this.triggerUserActing(isTyping, 'text', channelExternalId);
  }

  async triggerUserRecordingVoice(
    isRecording: boolean,
    channelExternalId: string
  ) {
    this.triggerUserActing(isRecording, 'voice', channelExternalId);
  }

  async triggerUserRecordingVideo(
    isRecording: boolean,
    channelExternalId: string
  ) {
    this.triggerUserActing(isRecording, 'video', channelExternalId);
  }

  async triggerUserActing(
    isActing: boolean,
    actingType: IActingType,
    channelExternalId: string
  ) {
    const channel = await this.getChannelByExternalId(channelExternalId);
    if (isActing) {
      const now = new Date();
      if (
        this.lastActingAt === null ||
        this.lastActingType !== actingType ||
        diffInSeconds(now, this.lastActingAt) > 5
      ) {
        this.lastActingAt = now;
        this.lastActingType = actingType;
        return this.chatClient.triggerUserActing(
          isActing ? new Date() : null,
          channel.id,
          actingType
        );
      }
    } else if (this.lastActingAt !== null) {
      this.lastActingAt = null;
      return this.chatClient.triggerUserActing(null, channel.id, actingType);
    }
  }

  markAllAsRead(channelId: string) {
    this.chatClient.emitMarkReadUntil(channelId);
  }

  async sendAttachment(
    channelId: string,
    messageType: ChatMessageType,
    content: string,
    file: any,
    replyTo?: ChatMessage | null,
    localId?: string,
    onUploadProgress?: (progress: number) => any
  ): Promise<void> {
    const currentUser = await this.getCurrentUser();
    // validate messageType
    const chatMessage: ChatMessage = {
      localId: localId || uuidv4(),
      channelId,
      messageType,
      content,
      postedAt: new Date(),
      createdAt: new Date().toString(),
      senderId: currentUser?.id,
      file,
      replyToId: replyTo?.id || null,
      replyTo,
    };
    // store the message as queued for retries
    this.channelMessageStorage(channelId).addLiveMessage([
      {
        ...chatMessage,
        id: chatMessage.localId,
        isQueued: true,
        isRetriable: this.channelsStorage.maxRetryAttemptsReached,
      },
    ]);
    return this.chatClient.sendAttachment(chatMessage, file, onUploadProgress);
  }

  async sendMessageExternally(
    channelExternalId: string,
    messageType: ChatMessageType,
    content: string,
    metadata?: Record<string, unknown>,
    replyToId?: string
  ) {
    const currentUser = await this.getCurrentUser();
    // validate messageType
    const chatMessage: ChatMessage = {
      localId: uuidv4(),
      channelExternalId,
      messageType,
      content,
      postedAt: new Date(),
      createdAt: new Date().toString(),
      senderId: currentUser?.id,
      metadata,
      replyToId: replyToId || null,
    };
    return this.chatClient.sendMessageExternally(chatMessage);
  }

  async loadEarlier(channelId: string) {
    const channelMessageStorage = this.channelMessageStorage(channelId);
    return channelMessageStorage.loadEarlier(this.chatClient);
  }

  async loadLater(channelId: string) {
    const channelMessageStorage = this.channelMessageStorage(channelId);
    return channelMessageStorage.loadLater(this.chatClient);
  }

  async searchMessages(channelId: string, query: string) {
    return this.chatClient.searchMessages(channelId, query);
  }

  async getUser(id: string) {
    const user = await this.chatClient.getUser(id);
    return user;
  }

  async blockUser(blockedId: string) {
    await this.chatClient.blockUser(blockedId);
    const blackList = await this.chatClient.getBlockedUsers();
    this.channelsStorage.setBlackList(blackList);
  }

  async unBlockUser(blockedId: string) {
    await this.chatClient.unBlockUser(blockedId);
    const blackList = await this.chatClient.getBlockedUsers();
    this.channelsStorage.setBlackList(blackList);
  }

  async archiveChannel(channelId: string) {
    await this.chatClient.archiveChannel(channelId);
    const archivedChannels = await this.chatClient.getArchivedChannels();
    this.channelsStorage.setArchivedChannels(archivedChannels, channelId);
  }

  async unArchiveChannel(channelId: string) {
    await this.chatClient.unArchiveChannel(channelId);
    const archivedChannels = await this.chatClient.getArchivedChannels();
    this.channelsStorage.setArchivedChannels(archivedChannels, channelId);
  }

  async draftMessage(channelId: string, draft: string) {
    await this.chatClient.updateDraftMessage(channelId, draft);
    const draftMessages = await this.chatClient.getDraftMessages();
    this.channelsStorage.setDraftMessages(draftMessages, channelId);
  }

  async deleteMessage(messageId: string) {
    return this.chatClient.deleteMessage(messageId);
  }

  loadEarlierChannels() {
    this.channelsStorage.loadEarlier(this.chatClient);
  }

  async getChannels(filter?: {
    skipEmpty?: boolean;
    metadata?: Record<any, any>;
    perPage?: number;
  }): Promise<ChatChannel[]> {
    return this.chatClient.getChannels(filter);
  }

  async getAllUnreadMessageCount(): Promise<{ unreadMessageCount: number }> {
    return this.chatClient.getAllUnreadMessageCount();
  }
}

export default ChatService;
