import { diffInSeconds, mergeDeep } from './utils';
import React, { useCallback } from 'react';

import { ChatMessage, ChatChannel } from './ChatClient';
import ChatService from './ChatService';

export type ChannelMessageStorageStateType = {
  messages: ChatMessage[];
  loadingEarlier: boolean;
  loadingLater: boolean;
  hasMoreEarlier: boolean;
  liveUpdate: boolean;
  channel?: ChatChannel;
  unreadMessageCount?: number;
};

type ReducerActionType = {
  type: 'update';
  channelId: string;
  chatService: ChatService;
  channel?: ChatChannel;
  lastReadMessageId?: number | null;
  unreadMessageCount?: number;
};

const initialState: ChannelMessageStorageStateType = {
  messages: [],
  loadingEarlier: false,
  loadingLater: false,
  hasMoreEarlier: false,
  liveUpdate: true,
  channel: undefined,
  unreadMessageCount: 0,
};

const reducer = (
  state: ChannelMessageStorageStateType,
  action: ReducerActionType
): ChannelMessageStorageStateType => {
  if (action.type === 'update') {
    const channelMessageStorage = action.chatService.channelMessageStorage(
      action.channelId
    );
    let data = { ...state, ...channelMessageStorage.state() };
    if (action.channel)
      data = {
        ...data,
        channel: mergeDeep({ ...state.channel }, { ...action.channel }),
      };
    if (action.unreadMessageCount !== undefined)
      data = { ...data, unreadMessageCount: action.unreadMessageCount };

    return data;
  }
  return state;
};

export const useChannelData = (
  channelExternalId: string,
  chatService: ChatService
): [
  ChannelMessageStorageStateType,
  () => Promise<void>,
  () => Promise<void>
] => {
  const [state, dispatch] = React.useReducer(reducer, initialState);
  const [currentUnsubscriptions, setCurrentUnsubscriptions] = React.useState<
    CallableFunction[]
  >([]);
  const loadEarlier = useCallback(async () => {
    if (state.channel) return chatService.loadEarlier(state.channel.id);
  }, [state.channel, chatService]);
  const loadLater = useCallback(async () => {
    if (state.channel) return chatService.loadLater(state.channel.id);
  }, [state.channel, chatService]);
  // let websocket listen to channel messages and subscribe to new new messages
  React.useEffect(() => {
    async function subscribeToChannel() {
      const channel = await chatService.getChannelByExternalId(
        channelExternalId
      );
      const channelMessageStorage = chatService.channelMessageStorage(
        channel.id
      );
      const removeListener = channelMessageStorage.onUpdate(() => {
        dispatch({
          type: 'update',
          channelId: channel.id,
          chatService,
        });
      });

      const unsubscribeChannel = chatService.subscribeChannel(channel.id);
      setCurrentUnsubscriptions([removeListener, unsubscribeChannel]);
    }

    if (
      state.channel?.externalId !== channelExternalId ||
      currentUnsubscriptions.length === 0
    ) {
      // unsubscribe from previous channels
      currentUnsubscriptions.forEach((cb) => cb());
      subscribeToChannel();
    }

    return () => {
      currentUnsubscriptions.forEach((cb) => cb());
      setCurrentUnsubscriptions([]);
    };
  }, [chatService, channelExternalId]);
  //fetch channel and listen for future channel updates
  React.useEffect(() => {
    const channelStorage = chatService.getChannelStorage();

    chatService
      .getChannelByExternalId(channelExternalId)
      .then(async (_channel) => {
        dispatch({
          type: 'update',
          channelId: _channel.id,
          unreadMessageCount: _channel.unreadMessageCount,
          chatService,
          channel: _channel,
        });
        chatService.setCurrentActiveChannel(_channel);
        chatService.markAllAsRead(_channel.id);
      })
      .catch((error) => {
        //TODO: Failed to load channel(due to server error)
        //Show error to user
        console.error('Failed Loading channel', error);
      });

    const removeListener = channelStorage.onUpdate((data: any) => {
      if (
        data &&
        data.channel &&
        data.channel.externalId === channelExternalId
      ) {
        dispatch({
          type: 'update',
          channelId: data.channel.id,
          chatService,
          channel: data.channel,
        });
        chatService.setCurrentActiveChannel(data.channel);
      }
    });
    return () => {
      chatService.setCurrentActiveChannel(null);
      removeListener();
    };
  }, [channelExternalId, chatService]);

  return [state, loadEarlier, loadLater];
};
