import React from 'react';

import { ChannelMembership, ChatUser, ActingUser } from './ChatClient';
import ChatService from './ChatService';
import { diffInSeconds } from './utils';

export type ChannelMembershipStoragStateType = {
  channelMemberships: ChannelMembership[];
  actingUsers?: ActingUser[];
};

type ReducerActionType = {
  type: 'update' | 'delete';
  chatService: ChatService;
  chatUser: ChatUser;
  actingUsers?: ActingUser[];
};

const initialState: ChannelMembershipStoragStateType = {
  channelMemberships: [],
  actingUsers: [],
};

const reducer = (
  state: ChannelMembershipStoragStateType,
  action: ReducerActionType
): ChannelMembershipStoragStateType => {
  const channelMembershipsStorage =
    action.chatService.getChannelMembershipsStorage();
  if (action.type === 'update') {
    return {
      ...state,
      ...channelMembershipsStorage.state(),
      actingUsers: action.actingUsers || state.actingUsers || [],
    };
  }
  return state;
};

export const useChannelMembershipsData = (
  chatService?: ChatService,
  currentUserId?: string,
  currentChannelId?: string
): {
  actingUsers: ActingUser[];
  allChannelMemberships: ChannelMembership[];
  currentUserMembership: ChannelMembership | undefined;
  otherUserMembership: ChannelMembership | undefined;
} => {
  const [state, dispatch] = React.useReducer(reducer, initialState);
  const currentUserMembership = React.useMemo(
    () =>
      state.channelMemberships.find(
        (channelMembership) =>
          channelMembership.userId === currentUserId &&
          channelMembership.channelId === currentChannelId
      ),
    [
      JSON.stringify(
        state.channelMemberships.find(
          (channelMembership) =>
            channelMembership.userId === currentUserId &&
            channelMembership.channelId === currentChannelId
        )
      ),
      currentUserId,
      currentChannelId,
    ]
  );
  const otherUserMembership = React.useMemo(() => {
    return state.channelMemberships.find(
      (channelMembership) =>
        currentUserId &&
        channelMembership.userId !== currentUserId &&
        channelMembership.channelId === currentChannelId
    );
  }, [
    JSON.stringify(
      state.channelMemberships.find(
        (channelMembership) =>
          currentUserId &&
          channelMembership.userId !== currentUserId &&
          channelMembership.channelId === currentChannelId
      )
    ),
    currentUserId,
    currentChannelId,
  ]);
  React.useEffect(() => {
    if (chatService) {
      const channelMembershipsStorage =
        chatService.getChannelMembershipsStorage();
      const removeUpdateListener = channelMembershipsStorage.onUpdate(
        async () => {
          const currentUser = await chatService.getCurrentUser();
          dispatch({
            type: 'update',
            chatService,
            chatUser: currentUser,
          });
        }
      );
      return () => {
        removeUpdateListener();
      };
    }
  }, [chatService]);
  React.useEffect(() => {
    async function checkTypingUsers(chatService: ChatService) {
      const currentUser = await chatService.getCurrentUser();
      const channelMembershipsStorage =
        chatService?.getChannelMembershipsStorage();
      const actingUsers = channelMembershipsStorage
        .sortedChannelMemberships()
        ?.filter(
          (memberShip) =>
            memberShip.userId !== currentUser?.id &&
            memberShip.lastActingAt &&
            diffInSeconds(new Date(), new Date(memberShip.lastActingAt)) < 20
        )
        .map((memberShip) => ({
          ...memberShip.user,
          actingType: memberShip.actingType,
        }));

      if (actingUsers && currentUser) {
        dispatch({
          type: 'update',
          actingUsers,
          chatService,
          chatUser: currentUser,
        });
      }
    }
    const intervalId = setInterval(() => {
      if (chatService) {
        checkTypingUsers(chatService);
      }
    }, 10 * 1000);
    if (chatService) {
      checkTypingUsers(chatService);
    }
    return () => {
      clearInterval(intervalId);
    };
  }, [
    state.channelMemberships?.filter(
      (memberShip) =>
        memberShip.lastActingAt &&
        diffInSeconds(new Date(), new Date(memberShip.lastActingAt)) < 20
    ).length,
    chatService,
  ]);

  return {
    actingUsers: state.actingUsers || [],
    allChannelMemberships: state.channelMemberships,
    currentUserMembership,
    otherUserMembership,
  };
};
