import React, { useCallback } from 'react';

import { BlockedUser, ChatChannel } from './ChatClient';
import ChatService from './ChatService';

export type ChannelStoragStateType = {
  channels: ChatChannel[];
  loadingEarlier: boolean;
  loadingLater: boolean;
  hasMoreEarlier: boolean;
  liveUpdate: boolean;
  errorMessage?: string;
  maxRetryAttemptsReached: boolean;
  blackList: BlockedUser[];
};

type ReducerActionType = {
  type: 'update' | 'delete';
  chatService: ChatService;
};

const initialState: ChannelStoragStateType = {
  channels: [],
  loadingEarlier: false,
  loadingLater: false,
  hasMoreEarlier: false,
  liveUpdate: true,
  errorMessage: undefined,
  maxRetryAttemptsReached: false,
  blackList: [],
};

const reducer = (
  state: ChannelStoragStateType,
  action: ReducerActionType
): ChannelStoragStateType => {
  if (action.type === 'update' || action.type === 'delete') {
    const channelsStorage = action.chatService.getChannelStorage();
    return { ...state, ...channelsStorage.state() };
  }
  return state;
};

export const useChannelsData = (
  chatService?: ChatService
): [ChannelStoragStateType, () => void] => {
  const [state, dispatch] = React.useReducer(reducer, initialState);
  const loadEarlier = useCallback(() => {
    if (chatService) {
      chatService.loadEarlierChannels();
    }
  }, [chatService]);

  React.useEffect(() => {
    if (chatService) {
      const channelsStorage = chatService.getChannelStorage();
      const removeUpdateListener = channelsStorage.onUpdate(() => {
        dispatch({ type: 'update', chatService });
      });
      const removeDeleteListener = channelsStorage.onDelete(() => {
        dispatch({ type: 'delete', chatService });
      });
      return () => {
        removeUpdateListener();
        removeDeleteListener();
      };
    }
  }, [chatService]);

  React.useEffect(() => {
    if (chatService) {
      const unSubscribeChannelsUpdate = chatService.subscribeChannelsUpdate();
      return () => {
        unSubscribeChannelsUpdate();
      };
    }
  }, [chatService]);

  return [state, loadEarlier];
};
